module Main where

import qualified Data.ByteString as B
import qualified Data.ByteString.Builder as BB
import qualified Data.ByteString.Char8 as B8
import Data.ByteString.UTF8 (fromString)
import qualified Data.Vector as V
import System.IO (stdout)
import Diff
import Diff3
import Format
import Hunks
import Merge
import Options.Applicative
import Redfa
import System.IO.MMap
import Version

data ADiffOptions =
  ADiffOptions
    { adiffRedfaOpt :: RedfaOption
    , adiffCmdOpts :: ADiffCommandOpts
    }
  deriving (Show)

data ADiffCommandOpts
  = CmdDiff
      { context :: Int
      , diffFile1 :: String
      , diffFile2 :: String
      }
  | CmdPatch
      { patchDryRun :: Bool
      , patchInDir :: Maybe String
      , patchInput :: String
      , patchReverse :: Bool
      , patchPathStrip :: Int
      , patchMergeOpts :: MergeOpts
      }
  | CmdDiff3
      { context :: Int
      , diff3Mine :: String
      , diff3Old :: String
      , diff3Yours :: String
      , diff3MergeOpts :: MergeOpts
      }
  deriving (Show)

diffCmdOptions =
  CmdDiff <$>
  option
    auto
    (metavar "CONTEXT" <>
     short 'C' <>
     long "context" <>
     value 5 <> help "How many tokens around the change to include in the patch") <*>
  strArgument (metavar "FROMFILE") <*>
  strArgument (metavar "TOFILE")

patchCmdOptions =
  CmdPatch <$>
  switch
    (short 'n' <>
     long "dry-run" <>
     help "Do not patch anything, just print what would happen") <*>
  optional
    (strOption $
     short 'd' <> long "directory" <> metavar "DIR" <> help "Work in DIR") <*>
  strOption
    (short 'i' <>
     long "input" <>
     metavar "INPUT" <>
     help "Read the patchfile from INPUT, defaults to `-' for STDIN" <>
     value "-") <*>
  switch (short 'R' <> long "reverse" <> help "Unapply applied patches") <*>
  option
    auto
    (short 'p' <>
     long "strip" <>
     metavar "NUM" <>
     help "Strip NUM leading components from the paths" <> value 0) <*>
  mergeOption True

diff3CmdOptions =
  CmdDiff3 <$>
  option
    auto
    (metavar "CONTEXT" <>
     short 'C' <>
     long "context" <>
     value 5 <> help "How many tokens around the change to include in the patch") <*>
  strArgument (metavar "MYFILE") <*>
  strArgument (metavar "OLDFILE") <*>
  strArgument (metavar "YOURFILE") <*>
  mergeOption False

actionOption :: Parser ADiffCommandOpts
actionOption =
  hsubparser $
  mconcat
    [ command "diff" $ info diffCmdOptions $ progDesc "Compare two files"
    , command "patch" $ info patchCmdOptions $ progDesc "Apply a patch to files"
    , command "diff3" $
      info diff3CmdOptions $ progDesc "Compare and merge three files"
    ]

adiffOptions = ADiffOptions <$> redfaOption <*> actionOption

main :: IO ()
main =
  let opts :: ParserInfo ADiffOptions
      opts =
        info
          (adiffOptions <**> versionOption "adiff" <**> helperOption)
          (fullDesc <>
           progDesc
             "Compare, patch and merge files on arbitrarily-tokenized sequences." <>
           header "adiff: arbitrary-token diff utilities")
   in do ADiffOptions {adiffRedfaOpt = ropt, adiffCmdOpts = copt} <-
           execParser opts
         redfa <- redfaPrepareRules ropt
         case copt of
           CmdDiff {diffFile1 = f1, diffFile2 = f2, context = ctxt} -> do
             data1 <- mmapFileByteString f1 Nothing
             data2 <- mmapFileByteString f2 Nothing
             toks1 <- V.fromList <$> redfaTokenize redfa data1
             toks2 <- V.fromList <$> redfaTokenize redfa data2
             B8.putStr $ pprHunks $ hunks (max 0 ctxt) $ diffToks toks1 toks2
           CmdPatch {} -> putStrLn "not supported yet"
           CmdDiff3 ctxt f1 f2 f3 mo -> do
             [toksMine, toksOld, toksYour] <-
               map V.fromList <$>
               traverse
                 ((>>= redfaTokenize redfa) . flip mmapFileByteString Nothing)
                 [f1, f2, f3]
             let d3 = diff3Toks toksMine toksOld toksYour
             if mergeDoMerge mo
               then BB.hPutBuilder stdout $ fmtMerged mo d3
               else B8.putStr $
                    pprHunks $
                    hunks (max 0 ctxt) $ diff3Toks toksMine toksOld toksYour
