{-# LANGUAGE FlexibleContexts #-}

module Redfa where

import Control.Monad
import qualified Data.ByteString as B
import qualified Data.ByteString.Builder as BB
import qualified Data.ByteString.Char8 as B8
import qualified Data.ByteString.Internal as BI
import qualified Data.ByteString.Lazy as BL
import Data.ByteString.UTF8 (fromString, toString)
import Data.List
import Data.Maybe
import Data.Monoid
import qualified Data.Vector as V
import Options.Applicative
import Text.Regex.TDFA
import Types
import Substr

data RedfaOption
  = RedfaOptionRules [BS]
  | RedfaOptionFile String
  deriving (Show)

data RedfaSpec =
  RedfaSpec
    { redfaStart :: Int
    , redfaNames :: V.Vector BS
    , redfaRules :: V.Vector [RedfaRule] --start, mid, jump, token
    }

data RedfaRule =
  RedfaRule
    { rrRegexStart :: Regex
    , rrRegexMid :: Regex
    , rrJump :: Int
    , rrIsToken :: Bool
    }

redfaOption :: Parser RedfaOption
redfaOption =
  RedfaOptionRules <$>
  some
    (strOption $
     metavar "RULE" <>
     short 'L' <>
     long "lex" <> help "Lexing rule (specify repeatedly for more rules)") <|>
  RedfaOptionFile <$>
  strOption
    (metavar "FILE" <>
     short 'l' <>
     long "lexer" <> help "File from where to load the lexing rules")

redfaOptionToRuleStrings :: RedfaOption -> IO [BS]
redfaOptionToRuleStrings (RedfaOptionRules x) = return x
redfaOptionToRuleStrings (RedfaOptionFile fn) =
  B8.lines <$> B.readFile fn -- TODO improve

splitFirst :: Char -> BS -> (BS, BS)
splitFirst c s = B.splitAt (fromMaybe (B.length s) $ B8.elemIndex c s) s

redfaRuleStringToRuleStr :: BS -> Maybe (BS, BS, BS, Bool)
redfaRuleStringToRuleStr s =
  let (spec, regex) = splitFirst ':' s
      (from, to) = splitFirst '>' spec
      sf = B8.strip from
      (cleanFrom, isToken)
        | B.null sf = (sf, True)
        | B.head sf == fromIntegral (fromEnum '_') = (B.tail sf, False)
        | otherwise = (sf, True)
      go
        | B.null s = Nothing
        | B.head s == 35 = Nothing -- # comment
        | B.null regex = Just (B.empty, B.empty, spec, isToken)
        | B.null to = Just (cleanFrom, cleanFrom, B.tail regex, isToken)
        | otherwise =
          Just (cleanFrom, B8.strip $ B.tail to, B.tail regex, isToken)
   in go

unescapeRegex :: MonadFail m => BS -> m BS
unescapeRegex s = BL.toStrict . BB.toLazyByteString <$> unescape' s
  where
    unescape' :: MonadFail m => BS -> m BB.Builder
    unescape' s
      | B.null s = return mempty
      | B.head s == BI.c2w '\\' && B.null (B.tail s) =
        fail "incomplete escape sequence"
      | B.head s == BI.c2w '\\' =
        let rest = B.tail s
            cc = B.head rest
            thechar =
              BB.stringUtf8 $
              case BI.w2c cc of
                '`' -> "\\`"
                '\'' -> "\\'"
                'b' -> "\\b"
                'B' -> "\\B"
                '<' -> "\\<"
                '>' -> "\\>"
                'a' -> "\a"
                'e' -> "\x1b"
                'f' -> "\f"
                'n' -> "\n"
                'r' -> "\r"
                't' -> "\t"
                '\\' -> "\\\\"
                a -> [a] --TODO add support for \x and \u
         in (thechar <>) <$> unescape' (B.tail rest)
      | otherwise = mappend (BB.word8 $ B.head s) <$> unescape' (B.tail s)

redfaPrepareRules :: RedfaOption -> IO RedfaSpec
redfaPrepareRules opt = do
  (states, jumps, regexes, isToken) <-
    unzip4 . mapMaybe redfaRuleStringToRuleStr <$>
    redfaOptionToRuleStrings opt
  uRegexes <- traverse unescapeRegex regexes
  startREs <- traverse (makeRegexM . (fromString "\\`" <>)) uRegexes
  midREs <- traverse (makeRegexM . (fromString "\\`(.|\n)" <>)) uRegexes
  let ids = nub . sort $ states ++ jumps
      index' = fromJust . flip elemIndex ids
      statesIds = map index' states
      jumpsIds = map index' jumps
      start = index' $ head states
      rules :: [[RedfaRule]]
      rules = do
        stateId <- [0 .. length ids - 1]
        return $ do
          (a, b, (rs, rm), t) <-
            zip4 statesIds jumpsIds (zip startREs midREs) isToken
          guard $ a == stateId
          return $ RedfaRule rs rm b t
  return $ RedfaSpec start (V.fromList ids) (V.fromList rules)

redfaTokenize :: MonadFail m => RedfaSpec -> BS -> m [Tok]
redfaTokenize spec s = redfaTokenize' spec s (redfaStart spec) 0 []

redfaTokenize' ::
     MonadFail m
  => RedfaSpec
  -> BS
  -> Int
  -> Int
  -> [Int]
  -> m [Tok]
redfaTokenize' spec s state off visited
  | off >= B.length s = pure []
  | otherwise =
    let (ooff, reg) =
          if off == 0
            then (0, rrRegexStart)
            else (1, rrRegexMid)
        matchString = B.drop (off - ooff) s
        matches :: [(RedfaRule, (MatchOffset, MatchLength))]
        matches =
          filter contOK $
          zip rules $ map (\x -> match (reg x) matchString) rules
        contOK (RedfaRule {rrJump = j}, (off', len))
          | off' /= 0 = False
          | len > ooff = True
          | otherwise = j `notElem` visited
     in case matches of
          [] ->
            fail $
            "Tokenization could not continue from " <>
            stateStrPretty <> " at offset " <> show (off + ooff)
          ((rule, (_, len)):_) ->
            let matchLen = len - ooff
             in (if matchLen > 0
                   then (:) (rrIsToken rule, substr off matchLen s)
                   else id) <$>
                redfaTokenize'
                  spec
                  s
                  (rrJump rule)
                  (off + matchLen)
                  (if matchLen > 0
                     then []
                     else state : visited)
  where
    rules = redfaRules spec V.! state
    stateStr = redfaNames spec V.! state
    stateStrPretty
      | B.null stateStr = "anonymous state"
      | otherwise = "state `" <> toString stateStr <> "'"
